from pkg_resources import get_distribution

from mup_tf.coord_check import get_coord_data, plot_coord_data
from mup_tf.infshape import InfDim, InfShape, zip_infshape
from mup_tf.layer import (
    MuOutConv1D,
    MuOutput,
    MuReadout,
    MuSharedReadout,
    reinit_layer,
    rescale_bias,
)
from mup_tf.optimizer import MuAdam, MuSGD
from mup_tf.shape import make_base_shapes, set_base_shapes

__version__ = get_distribution("mup_tf").version
__author__ = "Zach Nussbaum"

__all__ = [
    "InfShape",
    "InfDim",
    "zip_infshape",
    "MuAdam",
    "MuOutput",
    "MuReadout",
    "MuSGD",
    "MuSharedReadout",
    "MuOutConv1D",
    "rescale_bias",
    "reinit_layer",
    "set_base_shapes",
    "make_base_shapes",
    "get_coord_data",
    "plot_coord_data",
]



from copy import copy


class InfDim:
    """A dimension with a base dimension, used for calculating μP scaling.

    An `InfDim` object is made up of 2 numbers: a dimension and a base
    dimension. If the base dimension is None, then this object represents a
    "finite", or "non-width" dimension. Otherwise, it represents an "infinite",
    or "width" dimension.
    """

    def __init__(self, base_dim, dim):
        self.base_dim = base_dim
        self.dim = dim

    def isinf(self):
        return self.base_dim is not None

    def width_mult(self):
        """Width multiplier used for calculating μP scaling.

        Returns:
            1 if this is a finite dimension, or the ratio of the dimension to base_dim otherwise.
        """
        if self.isinf():
            if self.dim is None:
                raise ValueError(
                    "InfDim.width_mult() called on infinite dimension with dim=None"
                )

            return self.dim / self.base_dim
        return 1

    def __repr__(self):
        return f"InfDim({self.base_dim}, {self.dim})"

    def __str__(self):
        if self.isinf():
            return repr(self)
        return f"FinDim({self.dim})"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, InfDim):
            return False
        return self.base_dim == other.base_dim and self.dim == other.dim


class InfShape(tuple):
    """A tuple of `InfDim`s.

    This is intended to be attached to each parameter tensor `p` as `p.infshape`
    in a tf.keras.models.Model.trainable_variables.
    """

    def __init__(self, *args, **kwargs):
        tuple.__init__(*args, **kwargs)
        for dim in self:
            if not isinstance(dim, InfDim):
                raise ValueError("Elements of InfShape needs to be of class InfDim")
        # set main to be the last dimension that is infinite
        # for inf x inf this is fanin
        # for inf x fin or fin x inf it's the unique inf dim
        # user can set this manually if necessary
        self.main_idx = self.main = None
        for i, dim in list(enumerate(self))[::-1]:
            if dim.isinf():
                self.main_idx = i
                self.main = dim
                break

    def fanin_fanout(self):
        assert len(self) >= 2, "fanin, fanout undefined for 1-dimensional weights"
        return self[1], self[0]

    def fanin_fanout_mult_ratio(self):
        """Returns the ratio of fanin to fanout width multipliers.

        Returns:
            The ratio of fanin to fanout width multipliers.
        """
        fanin, fanout = self.fanin_fanout()
        return fanin.width_mult() / fanout.width_mult()

    def ninf(self):
        return sum(1 for dim in self if dim.isinf())

    def width_mult(self):
        if self.main is not None:
            return self.main.width_mult()
        return 1

    def base_shape(self):
        return [d.base_dim for d in self]

    def shape(self):
        return [d.dim for d in self]

    def __repr__(self):
        r = tuple.__repr__(self)[1:-1]
        return f"InfShape([{r}])"

    def serialize(self):
        d = {"base_shape": [], "shape": []}
        for infdim in self:
            d["shape"].append(infdim.dim)
            d["base_shape"].append(infdim.base_dim)
        return d

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, InfShape):
            return False
        return all(d == dd for d, dd in zip(self, other))

    @classmethod
    def deserialize(cls, d):
        infshape = []
        for base_dim, dim in zip(d["base_shape"], d["shape"]):
            infshape.append(InfDim(base_dim, dim))
        return InfShape(infshape)

    @classmethod
    def from_base_shape(cls, bsh):
        return InfShape([InfDim(bd, None) for bd in bsh])


def zip_infshape(base_dims, dims, fin_if_same=True):
    """Creates an `InfShape` from a list of shapes

    Args:
        base_dims: list of base dimensions. Can be of type `int` or `InfDim`
        dims: list of dimensions. Can be of type `int` or `InfDim`
        fin_if_same: If a dim in base_dims is the same as a dim in dims, then
            the created `InfDim` is treated as a finite dimension.

    Returns:
        An `InfShape` object.
    """
    infshape = []
    for bd, d in zip(base_dims, dims):
        if isinstance(bd, InfDim):
            # retain bd's base_dim but overwrite dim
            infdim = copy(bd)
            infdim.dim = d
            infshape.append(infdim)
        elif isinstance(bd, int):
            if bd == d and fin_if_same:
                infshape.append(InfDim(None, d))
            else:
                infshape.append(InfDim(bd, d))
        else:
            raise ValueError(f"unhandled base_dim type: {type(bd)}")
    return InfShape(infshape)



import tensorflow as tf


class MuAdam(tf.keras.optimizers.Adam):
    """Adam optimizer with μP-aware learning rate scaling.

    This behaves like the tf.keras.optimizers.Adam optimizer, but will scale
    the learning rate by the width multiplier.
    """

    def __init__(self, *args, **kwargs):
        self.infshapes = kwargs.pop("infshapes", None)
        super().__init__(*args, **kwargs)

    def _resource_apply_dense(self, grad, var, apply_state=None):
        # changed from https://github.com/keras-team/keras/blob/v2.9.0/keras/optimizers/optimizer_v2/adam.py#L198
        var_device, var_dtype = var.device, var.dtype.base_dtype
        coefficients = (apply_state or {}).get(
            (var_device, var_dtype)
        ) or self._fallback_apply_state(var_device, var_dtype)

        m = self.get_slot(var, "m")
        v = self.get_slot(var, "v")

        if hasattr(var, "infshape"):
            infshape = var.infshape
        else:
            infshape = self.infshapes[var.name]

        ninf = infshape.ninf()
        lr = coefficients["lr_t"]

        # from table 8 in https://arxiv.org/pdf/2203.03466.pdf
        if ninf == 2:
            lr /= infshape.width_mult()

        if not self.amsgrad:
            return tf.raw_ops.ResourceApplyAdam(
                var=var.handle,
                m=m.handle,
                v=v.handle,
                beta1_power=coefficients["beta_1_power"],
                beta2_power=coefficients["beta_2_power"],
                lr=lr,
                beta1=coefficients["beta_1_t"],
                beta2=coefficients["beta_2_t"],
                epsilon=coefficients["epsilon"],
                grad=grad,
                use_locking=self._use_locking,
            )
        else:
            vhat = self.get_slot(var, "vhat")
            return tf.raw_ops.ResourceApplyAdamWithAmsgrad(
                var=var.handle,
                m=m.handle,
                v=v.handle,
                vhat=vhat.handle,
                beta1_power=coefficients["beta_1_power"],
                beta2_power=coefficients["beta_2_power"],
                lr=lr,
                beta1=coefficients["beta_1_t"],
                beta2=coefficients["beta_2_t"],
                epsilon=coefficients["epsilon"],
                grad=grad,
                use_locking=self._use_locking,
            )


class MuSGD(tf.keras.optimizers.SGD):
    """Adam optimizer with μP-aware learning rate scaling.

    This behaves like the tf.keras.optimizers.SGD optimizer, but will scale
    the learning rate by the width multiplier.
    """

    def __init__(self, *args, **kwargs):
        self.infshapes = kwargs.pop("infshapes", None)
        super().__init__(*args, **kwargs)

    def _resource_apply_dense(self, grad, var, apply_state=None):
        # changed from https://github.com/keras-team/keras/blob/v2.9.0/keras/optimizers/optimizer_v2/gradient_descent.py#L132
        var_device, var_dtype = var.device, var.dtype.base_dtype
        coefficients = (apply_state or {}).get(
            (var_device, var_dtype)
        ) or self._fallback_apply_state(var_device, var_dtype)

        if hasattr(var, "infshape"):
            infshape = var.infshape
        else:
            infshape = self.infshapes[var.name]

        ninf = infshape.ninf()
        lr = coefficients["lr_t"]

        # from table 8 in https://arxiv.org/pdf/2203.03466.pdf
        if ninf == 1:
            lr *= infshape.width_mult()
        if ninf == 2:
            lr /= infshape.fanin_fanout_mult_ratio()

        if self._momentum:
            momentum_var = self.get_slot(var, "momentum")
            return tf.raw_ops.ResourceApplyKerasMomentum(
                var=var.handle,
                accum=momentum_var.handle,
                lr=lr,
                grad=grad,
                momentum=coefficients["momentum"],
                use_locking=self._use_locking,
                use_nesterov=self.nesterov,
            )
        else:
            return tf.raw_ops.ResourceApplyGradientDescent(
                var=var.handle, alpha=lr, delta=grad, use_locking=self._use_locking
            )



import tensorflow as tf


class MuOutput(tf.keras.layers.Layer):
    """Prototype for all output linear layers.
    An "output" linear layer is one that maps from a width dimension (e.g.,
    `d_model` in a Transformer) to a non-width dimension (e.g., vocab size).
    This layer implements the version of μP with a 1/width multiplier and a
    constant variance initialization for both weights and biases.
    """

    def __init__(
        self,
        *args,
        readout_zero_init=False,
        output_mult=1.0,
        kernel_initializer="he_uniform",
        bias_initializer="random_normal",
        **kwargs
    ):
        self.output_mult = output_mult
        self.readout_zero_init = readout_zero_init
        if self.readout_zero_init:
            kernel_initializer = tf.keras.initializers.Constant(0)
            bias_initializer = tf.keras.initializers.Constant(0)

        super().__init__(
            *args,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            **kwargs
        )

    def reset_parameters(self) -> None:
        if self.readout_zero_init:
            self.kernel.assign(tf.zeros_like(self.kernel))
            if self.bias is not None:
                self.bias.assign(tf.zeros_like(self.bias))
        else:
            if isinstance(self.kernel.kernel_initializer, str):
                self.kernel.kernel_initializer = tf.keras.initializers.get(
                    self.kernel.kernel_initializer
                )
            if isinstance(self.bias.bias_initializer, str):
                self.bias.bias_initializer = tf.keras.initializers.get(
                    self.bias.bias_initializer
                )

            self.kernel.assign(
                self.kernel.kernel_initializer(self.kernel.shape, self.kernel.dtype)
            )
            self.bias.assign(
                self.bias.bias_initializer(self.bias.shape, self.bias.dtype)
            )

    def width_mult(self):
        """Calculates the width multiplier for this layer.

        NOTE: We add a hack to return 1 if the layer has not been built yet (or just built)
        since `.build()` is called before `.set_base_shapes()`. To create variables
        in Tensorflow, you must call `.build()` first, otherwise there will no be `.kernel`
        attributed.  Internally, `build` calls `.call()` which calls `.width_mult()`.
        So we need to make sure that the model can be built before calling `.set_base_shapes()`.

        Returns:
            The width multiplier for this layer.
        """
        if hasattr(self.kernel, "infshape"):
            return self.kernel.infshape.width_mult()
        elif hasattr(self, "_has_rescaled_params"):
            raise ValueError("No infshape found for this layer.")
        else:
            # TODO this is a hack so when a layer is built, it can still call
            # Should we raise a warning? It's not easy to check if a layer was
            # just built or not.
            return 1

    def _rescale_parameters(self):
        """Rescale parameters to convert SP initialization to μP initialization.
        Warning: This method is NOT idempotent and should be called only once
        unless you know what you are doing.
        """
        if hasattr(self, "_has_rescaled_params") and self._has_rescaled_params:
            raise RuntimeError(
                "`_rescale_parameters` has been called once before already. "
                "Unless you know what you are doing, usually you should not be calling `_rescale_parameters` more than once.\n"
                "If you called `set_base_shapes` on a model loaded from a checkpoint, "
                "or just want to re-set the base shapes of an existing model, "
                "make sure to set the flag `rescale_params=False`.\n"
                "To bypass this error and *still rescale parameters*, set `self._has_rescaled_params=False` before this call."
            )
        if self.bias is not None:
            self.bias.assign(self.bias * self.width_mult() ** 0.5)
        self.kernel.assign(self.kernel * self.width_mult() ** 0.5)
        self._has_rescaled_params = True

    def call(self, x):
        output = super().call(self.output_mult * x / self.width_mult())
        return output


class MuReadout(MuOutput, tf.keras.layers.Dense):
    """Drop-in replacement for all output Linear layers."""

    def __init__(self, *args, readout_zero_init=False, output_mult=1.0, **kwargs):
        super().__init__(
            *args,
            readout_zero_init=readout_zero_init,
            output_mult=output_mult,
            **kwargs
        )


class MuSharedReadout(MuReadout):
    """`MuReadout` with weights shared with an `tf.keras.layers.Embedding` layer."""

    def __init__(self, weight, bias=True, **kwargs):
        """Args:
        weight: should be weight of an `tf.keras.layers.Embedding` layer
        """
        super().__init__(*weight.shape, bias=bias, **kwargs)
        self.weight = weight


class MuOutConv1D(MuOutput, tf.keras.layers.Conv1D):
    """Drop-in replacement for all output Conv1d layers."""

    def __init__(self, *args, readout_zero_init=False, output_mult=1.0, **kwargs):
        super().__init__(
            *args,
            readout_zero_init=readout_zero_init,
            output_mult=output_mult,
            **kwargs
        )


def rescale_bias(linear):
    """Rescale bias in tf.keras.layers.Dense or tf.keras.layers.ConvND layers to convert SP initialization to μP initialization.
    Warning: This method is NOT idempotent and should be called only once
    unless you know what you are doing.

    Args:
        linear: a tf.keras.layers.Dense or tf.keras.layers.ConvND layer
    """
    if hasattr(linear, "_has_rescaled_params") and linear._has_rescaled_params:
        raise RuntimeError(
            "`rescale_linear_bias` has been called once before already. Unless you know what you are doing, usually you should not be calling `rescale_linear_bias` more than once.\n"
            "If you called `set_base_shapes` on a model loaded from a checkpoint, or just want to re-set the base shapes of an existing model, make sure to set the flag `rescale_params=False`.\n"
            "To bypass this error and *still rescale biases*, set `linear._has_rescaled_params=False` before this call."
        )
    if linear.bias is None:
        return
    fanin_mult = linear.trainable_weights[0].infshape[1].width_mult()
    linear.bias = linear.bias * fanin_mult**0.5
    linear._has_rescaled_params = True


def reinit_layer(
    layer, kernel_initializer="he_uniform", bias_initializer="random_normal"
):
    """Reinitializes layers to have μP variance

    Args:
        layer: a tf.keras.layers.Dense or tf.keras.layers.ConvND layer
        kernel_initializer: initializer for kernel (str or tf.keras.initializers.Initializer)
        bias_initializer: initializer for bias (str or tf.keras.initializers.Initializer)
    """
    if (
        layer.kernel_initializer == kernel_initializer
        and layer.bias_initializer == bias_initializer
    ):
        return layer

    if isinstance(kernel_initializer, str):
        kernel_initializer = tf.keras.initializers.get(kernel_initializer)
    if isinstance(bias_initializer, str):
        bias_initializer = tf.keras.initializers.get(bias_initializer)

    layer.kernel_initializer = kernel_initializer
    layer.bias_initializer = bias_initializer

    if layer.built:
        # Reset weights and biases if layer is already built
        layer.kernel.assign(layer.kernel_initializer(layer.kernel.shape, layer.dtype))
        if layer.bias is not None:
            layer.bias.assign(layer.bias_initializer(layer.bias.shape, layer.dtype))

    return layer


def extract_layers(seq_layers, extract_custom=True):
    """Extract layers from a sequential layer.

    Args:
        seq_layers: a tf.keras.layers.Sequential

    Returns:
        a list of layers from the sequential layer.
    """
    layers = []
    for layer in seq_layers.layers:
        if isinstance(layer, tf.keras.Sequential):
            layers.extend(extract_layers(layer, extract_custom=extract_custom))
        # custom layers track their sublayers in `obj.layers`
        elif extract_custom and hasattr(layer, "layers"):
            layers.extend(extract_layers(layer, extract_custom=extract_custom))
        elif not layer.trainable or isinstance(
            layer, (tf.keras.layers.Activation, tf.keras.layers.Dropout)
        ):
            continue
        else:
            layers.append(layer)
    return layers
